@@ -1,31 +1,35 @@
 # This file is part of Xpra.
-# Copyright (C) 2010-2018 Antoine Martin <antoine@devloop.org.uk>
+# Copyright (C) 2010-2022 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
-from xpra.log import Logger
-avsynclog = Logger("av-sync")
-log = Logger("client", "sound")
+from gi.repository import GLib  # @UnresolvedImport
 
 from xpra.platform.paths import get_icon_filename
 from xpra.scripts.parsing import sound_option
 from xpra.net.compression import Compressed
+from xpra.net.protocol import CONNECTION_LOST
 from xpra.os_util import get_machine_id, get_user_uuid, bytestostr, OSX, POSIX
-from xpra.util import envint, typedict, csv
+from xpra.util import envint, typedict, csv, updict
 from xpra.client.mixins.stub_client_mixin import StubClientMixin
+from xpra.log import Logger
 
+avsynclog = Logger("av-sync")
+log = Logger("client", "sound")
 
 AV_SYNC_DELTA = envint("XPRA_AV_SYNC_DELTA")
+DELTA_THRESHOLD = envint("XPRA_AV_SYNC_DELTA_THRESHOLD", 40)
+DEFAULT_AV_SYNC_DELAY = envint("XPRA_DEFAULT_AV_SYNC_DELAY", 150)
 
 
-"""
-Utility superclass for clients that handle audio
-"""
 class AudioClient(StubClientMixin):
+    """
+    Utility mixin for clients that handle audio
+    """
     __signals__ = ["speaker-changed", "microphone-changed"]
 
     def __init__(self):
-        StubClientMixin.__init__(self)
+        super().__init__()
         self.sound_source_plugin = None
         self.speaker_allowed = False
         self.speaker_enabled = False
@@ -39,9 +43,10 @@ class AudioClient(StubClientMixin):
         #sound state:
         self.on_sink_ready = None
         self.sound_sink = None
-        self.min_sound_sequence = 0
+        self.sound_sink_sequence = 0
         self.server_sound_eos_sequence = False
         self.sound_source = None
+        self.sound_source_sequence = 0
         self.sound_in_bytecount = 0
         self.sound_out_bytecount = 0
         self.server_av_sync = False
@@ -51,9 +56,9 @@ class AudioClient(StubClientMixin):
         self.server_sound_encoders = []
         self.server_sound_receive = False
         self.server_sound_send = False
-        self.server_sound_bundle_metadata = False
-        self.server_ogg_latency_fix = False
         self.queue_used_sent = None
+        #duplicated from ServerInfo mixin:
+        self._remote_machine_id = None
 
     def init(self, opts):
         self.av_sync = opts.av_sync
@@ -70,11 +75,12 @@ class AudioClient(StubClientMixin):
             return []
         if self.speaker_allowed or self.microphone_allowed:
             try:
-                from xpra import sound
-                assert sound
+                from xpra.sound import common
+                assert common
             except ImportError as e:
-                log.warn("Warning: sound module is not installed")
-                log.warn(" speaker and microphone are disabled")
+                self.may_notify_audio("No Audio",
+                                      "audio subsystem is not installed\n" +
+                                      " speaker and microphone forwarding are disabled")
                 self.speaker_allowed = False
                 self.microphone_allowed = False
             else:
@@ -83,20 +89,19 @@ class AudioClient(StubClientMixin):
                     from xpra.sound.wrapper import query_sound
                     self.sound_properties = query_sound()
                     assert self.sound_properties, "query did not return any data"
-                    def vinfo(k):
-                        val = self.sound_properties.strlistget(k)
-                        assert val, "%s not found in sound properties" % k
-                        return ".".join(val[:3])
-                    bits = self.sound_properties.intget("python.bits", 32)
-                    log.info("GStreamer version %s for Python %s %s-bit", vinfo("gst.version"), vinfo("python.version"), bits)
+                    gstv = self.sound_properties.strtupleget("gst.version")
+                    if gstv:
+                        log.info("GStreamer version %s", ".".join(gstv[:3]))
+                    else:
+                        log.info("GStreamer loaded")
                 except Exception as e:
                     log("failed to query sound", exc_info=True)
                     log.error("Error: failed to query sound subsystem:")
                     log.error(" %s", e)
                     self.speaker_allowed = False
                     self.microphone_allowed = False
-        encoders = self.sound_properties.strlistget("encoders", [])
-        decoders = self.sound_properties.strlistget("decoders", [])
+        encoders = self.sound_properties.strtupleget("encoders")
+        decoders = self.sound_properties.strtupleget("decoders")
         self.speaker_codecs = sound_option_or_all("speaker-codec", opts.speaker_codec, decoders)
         self.microphone_codecs = sound_option_or_all("microphone-codec", opts.microphone_codec, encoders)
         if not self.speaker_codecs:
@@ -105,9 +110,9 @@ class AudioClient(StubClientMixin):
             self.microphone_allowed = False
         self.speaker_enabled = self.speaker_allowed and sound_option(opts.speaker)=="on"
         self.microphone_enabled = self.microphone_allowed and opts.microphone.lower()=="on"
-        self.av_sync = opts.av_sync
         log("speaker: codecs=%s, allowed=%s, enabled=%s", encoders, self.speaker_allowed, csv(self.speaker_codecs))
-        log("microphone: codecs=%s, allowed=%s, enabled=%s, default device=%s", decoders, self.microphone_allowed, csv(self.microphone_codecs), self.microphone_device)
+        log("microphone: codecs=%s, allowed=%s, enabled=%s, default device=%s",
+            decoders, self.microphone_allowed, csv(self.microphone_codecs), self.microphone_device)
         log("av-sync=%s", self.av_sync)
         if POSIX and not OSX:
             try:
@@ -120,6 +125,8 @@ class AudioClient(StubClientMixin):
                 log.warn(" %s", e)
             except Exception:
                 log.error("failed to add pulseaudio info", exc_info=True)
+        #audio tagging:
+        self.init_audio_tagging(opts.tray_icon)
 
 
     def cleanup(self):
@@ -132,7 +139,35 @@ class AudioClient(StubClientMixin):
         if self.sound_sink:
             self.stop_receiving_sound()
 
-    def get_audio_capabilities(self):
+
+    def get_info(self) -> dict:
+        info = {
+            "speaker" : self.speaker_enabled,
+            "microphone" : self.microphone_enabled,
+            "properties" : dict(self.sound_properties),
+            }
+        ss = self.sound_source
+        if ss:
+            info["src"] = ss.get_info()
+        ss = self.sound_sink
+        if ss:
+            info["sink"] = ss.get_info()
+        return {"audio" : info}
+
+
+    def get_caps(self) -> dict:
+        d = {}
+        avcaps = self.get_avsync_capabilities()
+        acaps = self.get_audio_capabilities()
+        #legacy flat format:
+        updict(d, "av-sync", avcaps)
+        updict(d, "sound", acaps)
+        #v4.4 namespace:
+        d["av-sync"] = avcaps
+        d["audio"] = acaps
+        return d
+
+    def get_audio_capabilities(self) -> dict:
         if not self.sound_properties:
             return {}
         #we don't know if the server supports new codec names,
@@ -148,32 +183,38 @@ class AudioClient(StubClientMixin):
         log("audio capabilities: %s", caps)
         return caps
 
-    def setup_connection(self, _conn):
-        pass
+    def get_avsync_capabilities(self) -> dict:
+        if not self.av_sync:
+            return {}
+        delay = max(0, DEFAULT_AV_SYNC_DELAY + AV_SYNC_DELTA)
+        return {
+            ""              : True,
+            "enabled"       : True,
+            "delay.default" : delay,
+            "delay"         : delay,
+            }
 
-    def parse_server_capabilities(self):
-        c = self.server_capabilities
+
+    def parse_server_capabilities(self, c : typedict) -> bool:
         self.server_av_sync = c.boolget("av-sync.enabled")
         avsynclog("av-sync: server=%s, client=%s", self.server_av_sync, self.av_sync)
         self.server_pulseaudio_id = c.strget("sound.pulseaudio.id")
         self.server_pulseaudio_server = c.strget("sound.pulseaudio.server")
-        try:
-            self.server_sound_decoders = c.strlistget("sound.decoders", [])
-            self.server_sound_encoders = c.strlistget("sound.encoders", [])
-        except:
-            log("Error: cannot parse server sound codec data", exc_info=True)
+        self.server_sound_decoders = c.strtupleget("sound.decoders")
+        self.server_sound_encoders = c.strtupleget("sound.encoders")
         self.server_sound_receive = c.boolget("sound.receive")
         self.server_sound_send = c.boolget("sound.send")
-        self.server_sound_bundle_metadata = c.boolget("sound.bundle-metadata")
-        self.server_ogg_latency_fix = c.boolget("sound.ogg-latency-fix", False)
         log("pulseaudio id=%s, server=%s, sound decoders=%s, sound encoders=%s, receive=%s, send=%s",
                  self.server_pulseaudio_id, self.server_pulseaudio_server,
                  csv(self.server_sound_decoders), csv(self.server_sound_encoders),
                  self.server_sound_receive, self.server_sound_send)
         if self.server_sound_send and self.speaker_enabled:
+            self.show_progress(90, "starting speaker forwarding")
             self.start_receiving_sound()
         if self.server_sound_receive and self.microphone_enabled:
-            self.start_sending_sound()
+            #call via idle_add because we may query X11 properties
+            #to find the pulseaudio server:
+            GLib.idle_add(self.start_sending_sound)
         return True
 
 
@@ -199,7 +240,7 @@ class AudioClient(StubClientMixin):
 
 
     def get_matching_codecs(self, local_codecs, server_codecs):
-        matching_codecs = [x for x in local_codecs if x in server_codecs]
+        matching_codecs = tuple(x for x in local_codecs if x in server_codecs)
         log("get_matching_codecs(%s, %s)=%s", local_codecs, server_codecs, matching_codecs)
         return matching_codecs
 
@@ -235,7 +276,7 @@ class AudioClient(StubClientMixin):
             log.warn(" %s", x)
         return False
 
-    def no_matching_codec_error(self, forwarding="speaker", server_codecs=[], client_codecs=[]):
+    def no_matching_codec_error(self, forwarding="speaker", server_codecs=(), client_codecs=()):
         summary = "Failed to start %s forwarding" % forwarding
         body = "No matching codecs between client and server"
         self.may_notify_audio(summary, body)
@@ -275,16 +316,19 @@ class AudioClient(StubClientMixin):
         #find the matching codecs:
         matching_codecs = self.get_matching_codecs(self.microphone_codecs, self.server_sound_decoders)
         log("start_sound_source(%s) matching codecs: %s", device, csv(matching_codecs))
-        if len(matching_codecs)==0:
+        if not matching_codecs:
             self.no_matching_codec_error("microphone", self.server_sound_decoders, self.microphone_codecs)
             return False
         try:
             from xpra.sound.wrapper import start_sending_sound
             plugins = self.sound_properties.get("plugins")
-            ss = start_sending_sound(plugins, self.sound_source_plugin, device or self.microphone_device, None, 1.0, False, matching_codecs, self.server_pulseaudio_server, self.server_pulseaudio_id)
+            ss = start_sending_sound(plugins, self.sound_source_plugin, device or self.microphone_device,
+                                     None, 1.0, False, matching_codecs,
+                                     self.server_pulseaudio_server, self.server_pulseaudio_id)
             if not ss:
                 return False
             self.sound_source = ss
+            ss.sequence = self.sound_source_sequence
             ss.connect("new-buffer", self.new_sound_buffer)
             ss.connect("state-changed", sound_source_state_changed)
             ss.connect("new-stream", self.new_stream)
@@ -320,10 +364,14 @@ class AudioClient(StubClientMixin):
             self.emit("microphone-changed")
         self.sound_source = None
         if ss is None:
-            log.warn("Warning: cannot stop sound source which has not been started")
+            log.warn("Warning: cannot stop audio capture which has not been started")
             return
         #tell the server to stop:
-        self.send("sound-data", ss.codec or "", "", {"end-of-stream" : True})
+        self.send("sound-data", ss.codec or "", "", {
+            "end-of-stream" : True,
+            "sequence"      : ss.sequence,
+            })
+        self.sound_source_sequence += 1
         ss.cleanup()
 
     def start_receiving_sound(self):
@@ -335,7 +383,7 @@ class AudioClient(StubClientMixin):
                 log("start_receiving_sound: we already have a sound sink")
                 enabled = True
                 return
-            elif not self.server_sound_send:
+            if not self.server_sound_send:
                 log.error("Error receiving sound: support not enabled on the server")
                 return
             if not self.audio_loop_check("speaker"):
@@ -343,13 +391,10 @@ class AudioClient(StubClientMixin):
             #choose a codec:
             matching_codecs = self.get_matching_codecs(self.speaker_codecs, self.server_sound_encoders)
             log("start_receiving_sound() matching codecs: %s", csv(matching_codecs))
-            if len(matching_codecs)==0:
+            if not matching_codecs:
                 self.no_matching_codec_error("speaker", self.server_sound_encoders, self.speaker_codecs)
                 return
             codec = matching_codecs[0]
-            if not self.server_ogg_latency_fix and codec in ("flac", "opus", "speex"):
-                log.warn("Warning: this server's sound support is out of date")
-                log.warn(" the sound latency with the %s codec will be high", codec)
             def sink_ready(*args):
                 scodec = codec
                 log("sink_ready(%s) codec=%s (server codec name=%s)", args, codec, scodec)
@@ -370,12 +415,12 @@ class AudioClient(StubClientMixin):
         if self.speaker_enabled:
             self.speaker_enabled = False
             self.emit("speaker-changed")
-        if tell_server:
-            self.send("sound-control", "stop", self.min_sound_sequence)
-        self.min_sound_sequence += 1
-        self.send("sound-control", "new-sequence", self.min_sound_sequence)
-        if ss is None:
+        if not ss:
             return
+        if tell_server and ss.sequence==self.sound_sink_sequence:
+            self.send("sound-control", "stop", self.sound_sink_sequence)
+        self.sound_sink_sequence += 1
+        self.send("sound-control", "new-sequence", self.sound_sink_sequence)
         self.sound_sink = None
         log("stop_receiving_sound(%s) calling %s", tell_server, ss.cleanup)
         ss.cleanup()
@@ -386,7 +431,7 @@ class AudioClient(StubClientMixin):
             log("sound_sink_state_changed(%s, %s) not the current sink, ignoring it", sound_sink, state)
             return
         log("sound_sink_state_changed(%s, %s) on_sink_ready=%s", sound_sink, state, self.on_sink_ready)
-        if state==b"ready" and self.on_sink_ready:
+        if bytestostr(state)=="ready" and self.on_sink_ready:
             if not self.on_sink_ready():
                 self.on_sink_ready = None
         self.emit("speaker-changed")
@@ -398,15 +443,17 @@ class AudioClient(StubClientMixin):
         #not shown in the UI, so don't bother with emitting a signal:
         #self.emit("speaker-changed")
     def sound_sink_error(self, sound_sink, error):
+        log("sound_sink_error(%s, %s) exit_code=%s, current sink=%s", sound_sink, error, self.exit_code, self.sound_sink)
         if self.exit_code is not None:
             #exiting
             return
         if sound_sink!=self.sound_sink:
             log("sound_sink_error(%s, %s) not the current sink, ignoring it", sound_sink, error)
             return
-        self.may_notify_audio("Speaker forwarding error", error)
+        estr = bytestostr(error).replace("gst-resource-error-quark: ", "")
+        self.may_notify_audio("Speaker forwarding error", estr)
         log.warn("Error: stopping speaker:")
-        log.warn(" %s", bytestostr(error).replace("gst-resource-error-quark: ", ""))
+        log.warn(" %s", estr)
         self.stop_receiving_sound()
     def sound_process_stopped(self, sound_sink, *args):
         if self.exit_code is not None:
@@ -443,12 +490,12 @@ class AudioClient(StubClientMixin):
             ss = start_receiving_sound(codec)
             if not ss:
                 return False
+            ss.sequence = self.sound_sink_sequence
             self.sound_sink = ss
             ss.connect("state-changed", self.sound_sink_state_changed)
             ss.connect("error", self.sound_sink_error)
             ss.connect("exit", self.sound_sink_exit)
-            from xpra.net.protocol import Protocol
-            ss.connect(Protocol.CONNECTION_LOST, self.sound_process_stopped)
+            ss.connect(CONNECTION_LOST, self.sound_process_stopped)
             ss.start()
             log("%s sound sink started", codec)
             return True
@@ -457,30 +504,24 @@ class AudioClient(StubClientMixin):
             self.sound_sink_error(self.sound_sink, e)
             return False
 
-    def new_sound_buffer(self, sound_source, data, metadata, packet_metadata=[]):
-        log("new_sound_buffer(%s, %s, %s, %s)", sound_source, len(data or []), metadata, packet_metadata)
-        if not self.sound_source:
+    def new_sound_buffer(self, sound_source, data, metadata, packet_metadata=()):
+        log("new_sound_buffer(%s, %s, %s, %s)", sound_source, len(data or ()), metadata, packet_metadata)
+        if sound_source.sequence<self.sound_source_sequence:
+            log("sound buffer dropped: old sequence number: %s (current is %s)",
+                sound_source.sequence, self.sound_source_sequence)
             return
         self.sound_out_bytecount += len(data)
         for x in packet_metadata:
             self.sound_out_bytecount += len(x)
+        metadata["sequence"] = sound_source.sequence
         if packet_metadata:
-            if not self.server_sound_bundle_metadata:
-                #server does not support bundling, send packet metadata as individual packets before the main packet:
-                for x in packet_metadata:
-                    self.send_sound_data(sound_source, x)
-                packet_metadata = ()
-            else:
-                #the packet metadata is compressed already:
-                packet_metadata = Compressed("packet metadata", packet_metadata, can_inline=True)
+            #the packet metadata is already compressed:
+            packet_metadata = Compressed("packet metadata", packet_metadata, can_inline=True)
         self.send_sound_data(sound_source, data, metadata, packet_metadata)
 
-    def send_sound_data(self, sound_source, data, metadata={}, packet_metadata=()):
+    def send_sound_data(self, sound_source, data, metadata, packet_metadata=None):
         codec = sound_source.codec
-        packet_data = [codec, Compressed(codec, data), metadata]
-        if packet_metadata:
-            assert self.server_sound_bundle_metadata
-            packet_data.append(packet_metadata)
+        packet_data = [codec, Compressed(codec, data), metadata, packet_metadata or ()]
         self.send("sound-data", *packet_data)
 
     def send_sound_sync(self, v):
@@ -497,8 +538,8 @@ class AudioClient(StubClientMixin):
             self.sound_in_bytecount += len(data)
         #verify sequence number if present:
         seq = metadata.intget("sequence", -1)
-        if self.min_sound_sequence>0 and seq>=0 and seq<self.min_sound_sequence:
-            log("ignoring sound data with old sequence number %s (now on %s)", seq, self.min_sound_sequence)
+        if self.sound_sink_sequence>0 and 0<=seq<self.sound_sink_sequence:
+            log("ignoring sound data with old sequence number %s (now on %s)", seq, self.sound_sink_sequence)
             return
 
         if not self.speaker_enabled:
@@ -532,7 +573,7 @@ class AudioClient(StubClientMixin):
             log.error(" stream tried to switch from %s to %s", ss.codec, codec)
             self.stop_receiving_sound()
             return
-        elif ss.get_state()=="stopped":
+        if ss.get_state()=="stopped":
             log("sound data received, sound sink is stopped - telling server to stop")
             self.stop_receiving_sound()
             return
@@ -540,23 +581,20 @@ class AudioClient(StubClientMixin):
         packet_metadata = ()
         if len(packet)>4:
             packet_metadata = packet[4]
-            if not self.sound_properties.get("bundle-metadata"):
-                #we don't handle bundling, so push individually:
-                for x in packet_metadata:
-                    ss.add_data(x)
-                packet_metadata = ()
         #(some packets (ie: sos, eos) only contain metadata)
-        if len(data)>0 or packet_metadata:
+        if data or packet_metadata:
             ss.add_data(data, metadata, packet_metadata)
         if self.av_sync and self.server_av_sync:
-            info = ss.get_info()
-            queue_used = info.get("queue.cur") or info.get("queue", {}).get("cur")
+            qinfo = typedict(ss.get_info()).dictget("queue")
+            queue_used = typedict(qinfo or {}).intget("cur", None)
             if queue_used is None:
                 return
             delta = (self.queue_used_sent or 0)-queue_used
-            #avsynclog("server sound sync: queue info=%s, last sent=%s, delta=%s", dict((k,v) for (k,v) in info.items() if k.startswith("queue")), self.queue_used_sent, delta)
-            if self.queue_used_sent is None or abs(delta)>=80:
-                avsynclog("server sound sync: sending updated queue.used=%i (was %s)", queue_used, (self.queue_used_sent or "unset"))
+            #avsynclog("server sound sync: queue info=%s, last sent=%s, delta=%s",
+            #    dict((k,v) for (k,v) in info.items() if k.startswith("queue")), self.queue_used_sent, delta)
+            if self.queue_used_sent is None or abs(delta)>=DELTA_THRESHOLD:
+                avsynclog("server sound sync: sending updated queue.used=%i (was %s)",
+                          queue_used, (self.queue_used_sent or "unset"))
                 self.queue_used_sent = queue_used
                 v = queue_used + self.av_sync_delta
                 if self.av_sync_delta:
@@ -567,6 +605,4 @@ class AudioClient(StubClientMixin):
     def init_authenticated_packet_handlers(self):
         log("init_authenticated_packet_handlers()")
         #these handlers can run directly from the network thread:
-        self.set_packet_handlers(self._packet_handlers, {
-            "sound-data":           self._process_sound_data,
-            })
+        self.add_packet_handler("sound-data", self._process_sound_data, False)
